"""
The `prodsys.factories` package contains the factory classes to create `prodsys.simulation` objects from `prodsys.models` objects in order to initialize a simulation run.

The following modules are available:

- `prodsys.factories.product_factory`: Contains the factory class to create products.
- `prodsys.factories.process_factory`: Contains the factory class to create processes.
- `prodsys.factories.queue_factory`: Contains the factory class to create queues.
- `prodsys.factories.resource_factory`: Contains the factory class to create resources.
- `prodsys.factories.sink_factory`: Contains the factory class to create sinks.
- `prodsys.factories.source_factory`: Contains the factory class to create sources.
- `prodsys.factories.state_factory`: Contains the factory class to create states.
- `prodsys.factories.time_model_factory`: Contains the factory class to create time models.

"""